package tdanford.sunlight;

import java.util.*;
import java.util.logging.Logger;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.*;
import java.net.URLEncoder;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

public class APIObject {

	public APIObject() {}
	public APIObject(JSONObject obj) { setFromJSONObject(obj); }
	
	public void setFromJSONObject(JSONObject obj) { 

		/*
		Logger logger = Logger.getLogger("APIObject");
		logger.entering(getClass().getSimpleName(),
				String.format("setFromJSONObject(%s)", obj.toString()));
		*/
		
		Class cls = getClass();
		Iterator<String> keys = obj.keys();
		while(keys.hasNext()) { 
			String key = keys.next();
			
			try {
				Object value = obj.get(key);
				Field f = cls.getField(key);
				int mod = f.getModifiers();
				if(!Modifier.isStatic(mod) && Modifier.isPublic(mod)) { 
					setFieldValue(this, new FieldSelector(f), value);
				}
				
			} catch (NoSuchFieldException e) {
				// do nothing.
			} catch (JSONException e) {
				throw new IllegalArgumentException(obj.toString(), e);
			}
		}
	}
	
	public Map<String,String> asParams() { 
		Map<String,String> paramMap = new LinkedHashMap<String,String>();
		
		for(Field f : getClass().getFields()) { 
			int mod = f.getModifiers();
			if(Modifier.isPublic(mod) && !Modifier.isStatic(mod)) { 
				try {
					Object value = f.get(this);
					if(value != null) {
						String paramKey = URLEncoder.encode(f.getName(), "UTF-8");
						String paramValue = URLEncoder.encode(String.valueOf(value), "UTF-8");
						paramMap.put(paramKey, paramValue);
					}
					
				} catch (IllegalAccessException e) {
					throw new IllegalStateException(e);
				} catch (UnsupportedEncodingException e) {
					throw new IllegalStateException(e);
				}
			}
		}
		
		return paramMap;
	}

	protected void setFieldValue(Object base, Selector field, Object value) {
		
		Class type = field.getType();
		try {
			if(type.isArray()) { 
				if(value != null && !(value instanceof JSONArray)) { 
					throw new IllegalArgumentException(String.format("%s : %s is not a JSONArray",
							String.valueOf(value), value.getClass().getSimpleName()));					
				}

				if(value == null) { 
					field.set(base, null);
				} else { 
					JSONArray array = (JSONArray)value;
					int length = array.length();
					Class arrayType = type.getComponentType();
					Object newArray = Array.newInstance(arrayType, length);

					for(int i = 0; i < length; i++) { 
						Selector sel = new ArraySelector(arrayType, i);
						setFieldValue(newArray, sel, array.get(i));
					}
					
					field.set(base, newArray);
				}
				
			} else if (isSubclass(type, APIObject.class)) { 

				if(value != null && !(value instanceof JSONObject)) { 
					throw new IllegalArgumentException(String.format("%s : %s is not a JSONObject",
							String.valueOf(value), value.getClass().getSimpleName()));										
				}
				
				if(value == null) { 
					field.set(base, null);
				} else { 
					Constructor<APIObject> constructor = type.getConstructor();
					APIObject obj = constructor.newInstance();
					obj.setFromJSONObject((JSONObject)value);
					field.set(base, obj);
				}

			} else if(isSubclass(type, String.class)) { 
				if(value != null && !(value instanceof String)) { 
					throw new IllegalArgumentException(String.format("%s : %s is not a String",
							String.valueOf(value), value.getClass().getSimpleName()));
				}

				String str = (String)value;
				field.set(base, str);

			} else if(isSubclass(type, Boolean.class)) { 
				if(value != null && !(value instanceof Boolean)) { 
					throw new IllegalArgumentException(String.format("%s : %s is not a Boolean",
							String.valueOf(value), value.getClass().getSimpleName()));
				}

				field.set(base, (Boolean)value);

			} else if (isSubclass(type, Integer.class)) { 
				if(value != null && !(value instanceof Integer) && !(value instanceof String)) { 
					throw new IllegalArgumentException(String.format("%s : %s is not an Integer or String",
							String.valueOf(value), value.getClass().getSimpleName()));
				}

				if(value instanceof Integer) { 
					Integer integer = (Integer)value;
					field.set(base, integer);
				} else if (value instanceof String) {
					try { 
						Integer integer = Integer.parseInt((String)value);
						field.set(base, integer);
					} catch(NumberFormatException e) { 
						throw new IllegalArgumentException(String.format("%s is not an integer", (String)value));
					}
				}

			} else if (isSubclass(type, Double.class)) { 
				if(value != null && !(value instanceof Double) && !(value instanceof String)) { 
					throw new IllegalArgumentException(String.format("%s : %s is not an Double or String",
							String.valueOf(value), value.getClass().getSimpleName()));
				}

				if(value instanceof Integer) { 
					Double dbl = (Double)value;
					field.set(base, dbl);
				} else if (value instanceof String) {
					try { 
						Double dbl = Double.parseDouble((String)value);
						field.set(base, dbl);
					} catch(NumberFormatException e) { 
						throw new IllegalArgumentException(String.format("%s is not an double", (String)value));
					}
				}

			} else { 
				throw new IllegalArgumentException(String.format("Field type %s cannot be handled.", type.getSimpleName()));
			}
			
		} catch (JSONException e) {
			throw new IllegalArgumentException(String.valueOf(value), e);
		} catch (NoSuchMethodException e) {
			throw new IllegalStateException(e);
		} catch (InstantiationException e) {
			throw new IllegalStateException(e);
		} catch (IllegalAccessException e) {
			throw new IllegalStateException(e);
		} catch (InvocationTargetException e) {
			throw new IllegalStateException(e);
		}
	}

	public static boolean isSubclass(Class<?> c1, Class<?> c2) {
		return c2.isAssignableFrom(c1);
	}
}

interface Selector<S,T> { 
	public void set(S base, T value);
	public T get(S base);
	public boolean isApplicable(S base);
	public Class<T> getType();
}

class ChainedSelectors<T,U,V> implements Selector<T,V> {
	
	private Selector<T,U> s1;
	private Selector<U,V> s2;
	
	public ChainedSelectors(Selector<T,U> s1, Selector<U,V> s2) { 
		this.s1 = s1;
		this.s2 = s2;
	}

	public V get(T base) {
		return s2.get(s1.get(base));
	}

	public Class<V> getType() {
		return s2.getType();
	}

	public boolean isApplicable(T base) {
		return s1.isApplicable(base) && s2.isApplicable(s1.get(base));
	}

	public void set(T base, V value) {
		s2.set(s1.get(base), value);
	}
	
	public String toString() { return String.format("%s%s", s1.toString(), s2.toString()); }
}

class ArraySelector<S,T> implements Selector<S, T>  {
	
	private int idx;
	private Class<T> type;
	
	public ArraySelector(Class<T> t, int i) {
		type = t;
		idx = i;
	}
	
	public Class<T> getType() { return type; }

	public T get(S base) {
		return (T)Array.get(base, idx); 
	}

	public boolean isApplicable(Object base) {
		if(base == null || !base.getClass().isArray()) { 
			return false;
		}
		
		int len = Array.getLength(base);
		return len > idx;
	}

	public void set(S base, T value) {
		Array.set(base, idx, value);
	}
	
	public String toString() { return String.format("[%d]", idx, type.getSimpleName()); }
}

class FieldSelector<S, T> implements Selector<S, T> {
	
	private Class<? extends S> baseClass;
	private Field field;
	private Class<T> type;
	
	public FieldSelector(S base, String fieldName) throws NoSuchFieldException {  
		this(base.getClass().getField(fieldName));
	}
	
	public FieldSelector(Field f) { 
		field = f;
		baseClass = (Class<? extends S>)f.getDeclaringClass();
		type = (Class<T>) field.getType();
	}
	
	public Class<T> getType() { return type; }

	public T get(S base) {
		try {
			return (T)field.get(base);
		} catch (IllegalAccessException e) {
			throw new IllegalStateException(field.toString(), e);
		}
	}

	public boolean isApplicable(Object base) {
		return APIObject.isSubclass(base.getClass(), baseClass);
	}

	public void set(S base, T value) {
		try {
			field.set(base, value);
		} catch (IllegalAccessException e) {
			throw new IllegalStateException(field.toString(), e);
		}
	}
	
	public String toString() { return String.format(".%s", field.getName()); }
}