package tdanford.sunlight;

import java.util.*;
import java.util.logging.Logger;
import java.util.regex.*;
import java.io.*;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.*;

import org.json.*;

public class SunlightLabsAPI implements API {
	
	public static void main(String[] args) throws IOException { 
		SunlightLabsAPI api = new SunlightLabsAPI();
		
		String zip = args[0];
		LegislatorArray response = api.legislators.allForZip(zip);
		
		for(int i = 0; i < response.legislators.length; i++) { 
			System.out.println(response.legislators[i]);
		}
	}
	
	private static String defaultURLBase = "http://services.sunlightlabs.com/api/";
	
	private String key;
	private String urlBase;
	private Logger logger;
	
	public LegislatorsAPI legislators = new LegislatorsAPI(this); 
	
	public SunlightLabsAPI() { 
		this(new APIProperties("sunlightlabs").getAPIKey());
	}
	
	public SunlightLabsAPI(String key) { 
		this(key, defaultURLBase);
	}
	
	public SunlightLabsAPI(String key, String urlBase) { 
		this.key = key;
		this.urlBase = urlBase;
		logger = Logger.getLogger("sunlightapi");
	}

	public <T extends APIObject> T request(Class<T> cls, String methodName, APIObject... params) throws IOException {
		try {
			LinkedHashMap<String,String> paramMap = new LinkedHashMap<String,String>();
			for(APIObject p : params) { 
				paramMap.putAll(p.asParams());
			}
			
			String responseString = rawRequest(methodName, paramMap);
			JSONObject obj = new JSONObject(responseString);
			
			logger.info(String.format("response: %s", obj.toString()));
			
			Constructor<T> constructor = cls.getConstructor();
			T response = constructor.newInstance();
			
			JSONObject inner = obj.getJSONObject("response");
			response.setFromJSONObject(inner);
		
			logger.info(String.format("inner: %s", inner.toString()));

			return response;
			
		} catch (NoSuchMethodException e) {
			throw new IllegalArgumentException(cls.getSimpleName(), e);
		} catch (InstantiationException e) {
			throw new IllegalArgumentException(cls.getSimpleName(), e);
		} catch (IllegalAccessException e) {
			throw new IllegalArgumentException(cls.getSimpleName(), e);
		} catch (InvocationTargetException e) {
			throw new IllegalArgumentException(cls.getSimpleName(), e);
		} catch (JSONException e) {
			throw new IOException(e);
		}
	}
	
	private String rawRequest(String methodName, Map<String,String> params) throws IOException {
		String url = urlBase + methodName + "?apikey=" + key;
		for(String p : params.keySet()) { 
			url += "&" + p + "=" + params.get(p);
		}

		logger.info(String.format("rawRequest URL: %s", url));
		
		try {
			URL Url = new URL(url);
			HttpURLConnection cxn = (HttpURLConnection) Url.openConnection();
			cxn.setRequestProperty("User-Agent", "Sunlight Labs Java API");
			
			int responseCode = cxn.getResponseCode();
			if(responseCode / 100 != 2) {
				logger.warning(String.format("Request returned response code %d (%s)", 
						responseCode, cxn.getResponseMessage()));
				throw new IOException(String.format("%d: %s", responseCode, cxn.getResponseMessage()));
			}
			
			InputStream is = cxn.getInputStream();
			BufferedReader br = new BufferedReader(new InputStreamReader(is));
			StringBuilder sb = new StringBuilder();
			String line;
			while((line = br.readLine()) != null) { 
				if(sb.length() > 0) { sb.append("\n"); }
				sb.append(line);
			}
			br.close();
			
			return sb.toString();

		} catch (MalformedURLException e) {
			logger.throwing(getClass().getSimpleName(), "rawRequest", e);
			throw new IllegalArgumentException(url);
		}
	}
}

class LegislatorsAPI extends ChildAPI { 
	
	public LegislatorsAPI(SunlightLabsAPI api) {
		super(api, "legislators");
	}

	public LegislatorArray allForZip(final String zipCode) throws IOException { 
		return request(LegislatorArray.class, "allForZip", new APIObject() { public String zip = zipCode; });
	}

	public LegislatorArray allForLatLong(final String lat, final String lng) throws IOException { 
		return request(LegislatorArray.class, "allForLatLong", 
				new APIObject() { public String latitude = lat, longitude = lng; });
	}

	public LegislatorSearchResultArray search(final String _name, final double _threshold, final boolean _all_legislators) throws IOException { 
		return search(new APIObject() { 
			public String name = _name; 
			public Double threshold = _threshold; 
			public Boolean all_legislators = _all_legislators;
		});
	}

	public LegislatorSearchResultArray search(APIObject params) throws IOException { 
		return request(LegislatorSearchResultArray.class, "search", params);
	}
	
	public static class LegislatorSearchResult extends APIObject { 
		public ScoredLegislatorEntry result;
	}
	
	public static class LegislatorSearchResultArray extends APIObject { 
		public LegislatorSearchResult[] results;
	}
}

class CommitteesAPI extends ChildAPI { 
	public CommitteesAPI(SunlightLabsAPI api) { 
		super(api, "committees");
	}
	
	public CommitteeEntry get(final String _id) throws IOException { 
		return request(CommitteeEntry.class, "get", new APIObject() { public String id = _id; }); 
	}
}


