package tdanford.sunlight.cgp;

import java.util.Collection;
import java.util.LinkedList;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

public class MARCXMLHandler extends DefaultHandler {
	
	private LinkedList<MARC21Record> recordList;
	private MARC21Record record;
	private MARC21Leader leader;
	private ControlField control;
	private DataField data;
	private SubField subfield;
	
	public MARCXMLHandler() {
		recordList = new LinkedList<MARC21Record>();
	}
	
	public Collection<MARC21Record> getRecords() { 
		return recordList;
	}
	
	public void startElement(String uri, String localName, String qName, Attributes atts) throws SAXException {
		String name = qName;
		
		if(name.equals("record")) {
			if(record != null) { throw new SAXException(); }
			
			record = new MARC21Record();
			
		} else if (name.equals("leader")) {
			if(record == null) { throw new SAXException(); }
			if(leader != null) { throw new SAXException(); }
			
			leader = new MARC21Leader();
			
		} else if (name.equals("controlfield")) {
			if(record == null) { throw new SAXException(); }
			if(control != null) { throw new SAXException(); }
			
			String tag = atts.getValue("tag");
			control = new ControlField(tag);
			
		} else if (name.equals("datafield")) { 
			if(record == null) { throw new SAXException(); }
			if(data != null) { throw new SAXException(); }

			String tag = atts.getValue("tag");
			char[] inds = parseIndicators(atts);
			data = new DataField(tag, inds);
			
		} else if (name.equals("subfield")) { 
			if(record == null) { throw new SAXException(); }
			if(data == null) { throw new SAXException(); }
			if(subfield != null) { throw new SAXException(); }
			
			char code = atts.getValue("code").charAt(0);
			subfield = new SubField(code);
		}
	}
	
	private char[] parseIndicators(Attributes attrs) { 
		TreeSet<String> indTags = new TreeSet<String>();
		
		Pattern p = Pattern.compile("ind\\d+");
		for(int i = 0; i < attrs.getLength(); i++) { 
			String key = attrs.getQName(i);
			Matcher keyMatcher = p.matcher(key);
			if(keyMatcher.matches()) { 
				indTags.add(key);
			}
		}
		
		char[] inds = new char[indTags.size()];
		int i = 0;
		for(String attrKey : indTags) { 
			inds[i++] = attrs.getValue(attrKey).charAt(0);
		}
		return inds;
	}

	public void endElement(String uri, String localName, String qName) throws SAXException {
		String name = qName;
		
		if(name.equals("record")) {
			if(record == null) { throw new SAXException(); }
			
			recordList.add(record);
			record = null;
			
		} else if (name.equals("leader")) {
			if(record == null) { throw new SAXException(); }			
			if(leader == null) { throw new SAXException(); }
			
			record.leader = leader;
			leader = null;
			
		} else if (name.equals("controlfield")) {
			if(record == null) { throw new SAXException(); }			
			if(control == null) { throw new SAXException(); }			
			
			record.controlFields.addLast(control);
			control = null;
			
		} else if (name.equals("datafield")) { 
			if(record == null) { throw new SAXException(); }			
			if(data == null) { throw new SAXException(); }
			
			record.dataFields.addLast(data);
			data = null;
			
		} else if (name.equals("subfield")) { 
			if(record == null) { throw new SAXException(); }			
			if(data == null) { throw new SAXException(); }			
			if(subfield == null) { throw new SAXException(); }			
			
			data.subfields.addLast(subfield);
			data.content += "\n";
			subfield = null;
		}
	}

	public void characters(char[] ch, int start, int length) throws SAXException {
		StringBuilder sb = new StringBuilder();
		for(int i = 0; i < length; i++) { 
			sb.append(ch[start+i]);
		}
		String content = sb.toString();
		
		if(leader != null) { 
			leader.content += content;
			
		} else if (subfield != null) {  // order matters; check subfield *before* data.
			subfield.content += content;
			
		} else if (control != null) { 
			control.content += content;			
		}
	}

	public void startDocument() throws SAXException {
		record = null;
		leader = null;
		control = null;
		data = null;
		subfield = null;
	}

	public void endDocument() throws SAXException {
	}
}
