package tdanford.sunlight.cgp;

import java.util.*;
import java.util.regex.*;
import java.io.*;

public class SunlightCGPDirectory {
	
	private static Pattern xmlFilePattern = Pattern.compile("^(\\d+)-(\\d+).xml$");

	private File dir;
	private Map<String,TreeMap<Integer,File>> versionedFiles;
	
	public SunlightCGPDirectory(File d) { 
		if(!d.isDirectory() || !d.canRead()) { 
			throw new IllegalArgumentException(String.format("%s is not a dir, or cannot read", 
					d.getAbsolutePath()));
		}
		
		dir = d;
		versionedFiles = new TreeMap<String,TreeMap<Integer,File>>();
		
		File[] list = dir.listFiles();
		for(File f : list) { 
			if(f.isDirectory()) { 
				SunlightCGPDirectory subdir = new SunlightCGPDirectory(f);
				String prefix = f.getName();
				
				for(String key : subdir.versionedFiles.keySet()) { 
					String newKey = String.format("%s-%s", prefix, key);
					versionedFiles.put(newKey, subdir.versionedFiles.get(key));
				}
				
			} else { 
				addFile(f);
			}
		}
	}
	
	private void addVersionedFile(String key, Integer version, File f) { 
		if(!versionedFiles.containsKey(key)) { versionedFiles.put(key, new TreeMap<Integer,File>()); }
		
		TreeMap<Integer,File> versions = versionedFiles.get(key);
		if(versions.containsKey(version) && !versions.get(version).equals(f)) { 
			throw new IllegalArgumentException(String.format("%s != %s for key %s, version %d", 
					versions.get(version), f, key, version));
		}
		
		versions.put(version, f);		
	}
	
	public void addFile(File f) { 
		Matcher m = xmlFilePattern.matcher(f.getName());
		if(m.matches()) { 
			
			String key = m.group(1);
			Integer version = Integer.parseInt(m.group(2));
	
			addVersionedFile(key, version, f);
		}
	}
	
	public Set<String> keys() { 
		return new TreeSet<String>(versionedFiles.keySet());
	}
	
	public File getLatestFile(String key) {
		if(!versionedFiles.containsKey(key)) { throw new IllegalArgumentException(key); }
		TreeMap<Integer,File> versions = versionedFiles.get(key);
		return versions.get(versions.lastKey());
	}
	
	public Collection<File> collectLatestFiles() { 
		LinkedList<File> latest = new LinkedList<File>();
		
		for(String key : versionedFiles.keySet()) { 
			TreeMap<Integer,File> versions = versionedFiles.get(key);
			Integer last = versions.lastKey();
			latest.add(versions.get(last));
		}
		
		return latest;
	}
}
